"""
Snaptree module. Groups the functions used to create the tree of checkpoints.
"""
import re
from collections import OrderedDict as OD
from datetime import datetime

from asciitree import LeftAligned


def convert_dt(creation_time: str) -> str:
    """
    Convert hyper-v timestamp to readable datetime.

    Args:
        creation_time: Timestamp with the snapshot creation date.
    Returns:
        Datetime string in readable format.
    """
    c = datetime.fromtimestamp(float(re.search("[0-9]+",
                                     creation_time).group())/1000.0)
    return c.strftime("%d/%m/%Y %H:%M:%S")


def create_tree(table: dict,
                root: str,
                mark: str,
                f_pid: str = "pid",
                f_id: str = "id",
                f_label: str = "item",
                f_ctime: str = "ctime",
                v_none: str = "null") -> str:
    """
    Creates the ascii checkpoint tree.

    Args:
        table: List of checkpoints as a table.
        root: Root label of the tree.
        mark: Current snapshot name. Mark the element with an '*'
        f_pid: Field of the table that contains the parent id of the node.
        f_id: Field of the table that contains the id of the node.
        f_label: Field with the label of the node.
        f_ctime: Field with the timestamp creation date.
        v_none: Value to be used for empty fields.
    Returns:
        A string containing the tree of snapshots
    """
    items = [(c[f_pid], c[f_id]) for c in table]
    tree = {root: OD()}
    inserted = []

    while [x[1] for x in items if x[1] not in inserted]:
        for parent, item in items:
            if parent == v_none:
                parent = root
            inserts = walk(tree, parent, item)
            inserted.extend(list(inserts))

    tr = LeftAligned()
    tr_tree = tr(tree)

    for cell in table:
        if cell[f_label] == mark:
            cell[f_label] = "{}*".format(cell[f_label])

        tr_tree = tr_tree.replace(cell[f_id],
                                  "{} ({})".format(
                                      cell[f_label],
                                      convert_dt(cell[f_ctime])))
    return tr_tree


def walk(node: dict, search_key: str, insert_key: str) -> str:
    """
    Walk in the tree to find where to insert the element.

    Args:
        node: The tree so far.
        search_key: Where to insert the new value.
        insert_key: The element to be inserted.
    Yields:
        List of inserted items in the current iteration.
    """
    for k, v in node.items():
        if k == search_key:
            if insert_key not in node[k].keys():
                node[k][insert_key] = OD()
                yield insert_key
        elif isinstance(v, OD):
            yield from walk(v, search_key, insert_key)



"""
Cache module. Interacts with the cache file.
"""
from datetime import datetime, timedelta
from json import JSONDecodeError, dump, load
from os import remove
from os.path import getmtime, isfile, join
from tempfile import gettempdir

sync_interval = None
current_host = None


def get_cache_path() -> str:
    """
    Get the cache file path based on the current Hyper-V host.

    Returns:
        The path to be used to create or update de cache file.
    """
    vms_cache = join(gettempdir(), 'vms_{}.cache'.format(current_host))
    return vms_cache


def get_vm_by_name(name: str) -> dict:
    """
    Get vm from cache.

    Args:
        name: The name of the virtual machine.
    Returns:
        The vm info found in the cache.
    """
    vms = list_vms()
    vm = [vm for vm in vms if vm['Name'] == name][0]
    vm['index'] = vms.index(vm)
    return vm


def get_vm_by_index(index: str) -> dict:
    """
    Get vm from cache.

    Args:
        index: The index of the vm in the cache.
    Returns:
        The vm info found in the cache.
    """
    vms = list_vms()
    vm = vms[int(index)]
    vm['index'] = index
    return vm


def list_vms() -> list:
    """
    Get all vms from cache.

    Returns:
        All vm info found in cache.
    """
    vms_cache_filename = get_cache_path()
    try:
        with open(vms_cache_filename, 'r') as vms_cache_file:
            vms_cache = load(vms_cache_file)
    except FileNotFoundError:
        print("Cache file not found")
        vms_cache = []
    except JSONDecodeError:
        print("Invalid cache file")
        vms_cache = []

    return vms_cache


def update_cache(vms_json: dict):
    """
    Update current cache with new information.

    Args:
        vms_json: Information to be included in cache.
    """
    # If there is only one element, make it a list
    if isinstance(vms_json, dict):
        vms_json = [vms_json]

    vms_cache_filename = get_cache_path()
    if isfile(vms_cache_filename):
        vms_cache = list_vms()
        vms_json = list({x['Id']: x for x in vms_cache + vms_json}.values())

    vms_json.sort(key=lambda k: k['Name'])
    with open(vms_cache_filename, 'w') as vms_cache_file:
        dump(vms_json, vms_cache_file, indent=4)


def need_update() -> bool:
    """
    Cheks if cache needs update based on cache file modification date.

    Returns:
        True if the cache file is older than sync interval in hours.
    """
    vms_cache_filename = get_cache_path()
    modified = datetime.min
    if isfile(vms_cache_filename):
        modified = datetime.fromtimestamp(getmtime(vms_cache_filename))

    if modified < datetime.now() - timedelta(hours=int(sync_interval)):
        return True

    return False


def remove_cache() -> bool:
    """
    Remove the current host cache file.

    Returns:
        Whether the file was removed with success.
    """
    vms_cache_filename = get_cache_path()
    if isfile(vms_cache_filename):
        remove(vms_cache_filename)

    return not isfile(vms_cache_filename)


def get_name(by_name, ident):
    """Retrieve name or name by index based on user input."""
    if by_name:
        name = ident
    else:
        index = ident
        name = get_vm_by_index(index)['Name']

    return name



import json
import platform
from base64 import b64encode
from collections import namedtuple
from subprocess import DEVNULL, PIPE, Popen, TimeoutExpired

from paramiko import AutoAddPolicy, SSHClient
from winrm import Protocol, Response

config = None

SNAP_TYPES = {'standard': 5,
              'production': 3,
              'productiononly': 4}


def connect(vm_id: str, vm_name: str, vm_index: str):
    """
    Connect to the virtual machine.

    Args:
        vm_name: Name of the vm to connect.
        vm_id: Hyper-v unique identificator of the vm.
        vm_index: Index of the vm in the cache file.
    """
    user = config['user']
    passw = config['pass']
    host = config['host']

    if platform.uname()[0] == "Windows":
        freerdp_bin = "wfreerdp.exe"
    else:
        freerdp_bin = "xfreerdp"

    cmd = [freerdp_bin, '/v:{}'.format(host),
                        '/vmconnect:{}'.format(vm_id),
                        '/u:{}'.format(user),
                        '/p:{}'.format(passw),
                        '/t:{} [{}] {}'.format(host, vm_index, vm_name),
                        '/cert-ignore']

    try:
        handle = Popen(cmd, stdout=DEVNULL, stderr=PIPE)
        errs = handle.communicate(timeout=5)[1]
        if errs:
            print(errs.decode('utf-8'))
    except FileNotFoundError as err:
        print("{} not found in PATH\n{}".format(freerdp_bin, err))
    except TimeoutExpired:
        print("{} started in background".format(freerdp_bin))


def get_vm(vm_name: str) -> Response:
    """
    Retrieve vm information from hyper-v.

    Args:
        vm_name: Name of the vm. Using * in the vm's name can retrieve info
            of one or more machines. If the vm_name is None, * will be used
            instead, gathering information about all vms in the host which can
            be slow depending on the number of vms.
    Returns:
        Info obtained from remote hyper-v host.
    """
    if not vm_name:
        vm_name = '*'
    ps_script = 'Get-VM -Name "{}" | Select Name,Id,State,Uptime,ParentSnapshotName | sort Name | \
ConvertTo-Json'.format(vm_name)
    rs = run_ps(ps_script)

    return rs


def list_vm_snaps(vm_name: str) -> Response:
    """
    List vm snapshots.

    Args:
        vm_name: The virtual machine name.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Get-VMSnapshot -VMName "{}" | Select Name,ParentSnapshotName,CreationTime,\
ParentSnapshotId,Id | ConvertTo-Json'.format(vm_name)

    rs = run_ps(ps_script)
    return rs


def restore_vm_snap(vm_name: str, snap_name: str) -> Response:
    """
    Restore virtual machine snapshot.

    Args:
        vm_name: The virtual machine name.
        snap_name: The name of the checkpoint to be restored.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Restore-VMSnapshot -Name "{}" -VMName "{}" -Confirm:$false'.format(snap_name,
                                                                                    vm_name)

    rs = run_ps(ps_script)
    return rs


def remove_vm_snapshot(vm_name: str, snap_name: str,
                       recursive: bool = False) -> Response:
    """
    Deletes a virtual machine checkpoint.

    Args:
        vm_name: The virtual machine name.
        snap_name: The name of the checkpoint to be deleted.
        recursive: Specifies that the checkpoint’s children
            are to be deleted along with the checkpoint.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Remove-VMSnapshot -VMName "{}" -Name "{}"'.format(vm_name,
                                                                   snap_name)
    if recursive:
        ps_script += " -IncludeAllChildSnapshots"
    ps_script += " -Confirm:$false"

    rs = run_ps(ps_script)
    return rs


def get_snapsshot_type(vm_name: str) -> Response:
    """
    Get snapshot type from vm.

    Args:
        vm_name: The virtual machine name.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Get-VM -VMName "{}" | Select CheckpointType | ConvertTo-Json'.format(vm_name)

    rs = run_ps(ps_script)
    return rs


def set_snapshot_type(vm_name: str, snap_type: str) -> Response:
    """
    Set snapshot type to be created in this vm.

    Args:
        vm_name: The virtual machine name.
        snap_type: The type of the checkpoint to be created.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Set-VM -VMName "{}" -CheckpointType {}'.format(vm_name, snap_type)

    rs = run_ps(ps_script)
    return rs


def create_vm_snapshot(vm_name: str, snap_name: str) -> Response:
    """
    Create a new snapshot with vm's current state.

    Args:
        vm_name: The virtual machine name.
        snap_name: The name of the checkpoint to be created.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Checkpoint-VM -Name "{}" -SnapshotName "{}" -Confirm:$false'.format(vm_name,
                                                                                     snap_name)

    rs = run_ps(ps_script)
    return rs


def parse_result(rs: Response) -> dict:
    """
    Parse Respnse object obtained from hyper-v.

    Args:
        rs: Response object with its properties (out, err and return_code).
    Returns:
        if there's something in std_out, return as json.
    """
    if rs.status_code != 0:
        print(rs.std_err)
        exit(1)

    if rs.std_out:
        rs_json = json.loads(rs.std_out.decode('latin-1'))
        return rs_json

    return None


def stop_vm(vm_name: str, force: bool = False) -> Response:
    """
    Stop virtual machine.

    Args:
        vm_name: The virtual machine name.
        force: Whether should force shutdown or not.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Stop-VM -Name "{}"'.format(vm_name)
    if force:
        ps_script += " -Force"

    rs = run_ps(ps_script)
    return rs


def resume_vm(vm_name: str) -> Response:
    """
    Resume (paused) virtual machine.

    Args:
        vm_name: The virtual machine name.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Resume-VM -Name "{}"'.format(vm_name)

    rs = run_ps(ps_script)
    return rs


def save_vm(vm_name: str) -> Response:
    """
    Save (Hibernate) virtual machine state.

    Args:
        vm_name: The virtual machine name.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Save-VM -Name "{}"'.format(vm_name)

    rs = run_ps(ps_script)
    return rs


def pause_vm(vm_name: str) -> Response:
    """
    Pause virtual machine.

    Args:
        vm_name: The virtual machine name.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Suspend-VM -Name "{}"'.format(vm_name)

    rs = run_ps(ps_script)
    return rs


def start_vm(vm_name: str) -> Response:
    """
    Start virtual machine.

    Args:
        vm_name: The virtual machine name.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Start-VM -Name "{}"'.format(vm_name)
    rs = run_ps(ps_script)

    return rs


def list_switches() -> Response:
    """
    List virtual network switches from server.

    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = "Get-VMSwitch | Select Name | ConvertTo-Json"
    rs = run_ps(ps_script)

    return rs


def get_switch(vm_name: str) -> Response:
    """
    Get current virtual network switch from mv.

    Args:
        vm_name: The virtual machine name.
    Returns:
        Info obtained from remote hyper-v host.
    """
    ps_script = 'Get-VMNetworkAdapter -VMName "{}" | Select VMName, SwitchName | ConvertTo-Json'\
        .format(vm_name)
    rs = run_ps(ps_script)

    return rs


def set_switch(vm_name: str, switch_name: str) -> Response:
    """
    Set virtual machine's virtual network switch.

    Args:
        vm_name: The virtual machine name.
        switch_name: Switch name as retrieved by get_switch.
    Returns:
        Info about the command execution.
    """
    ps_script = 'Connect-VMNetworkAdapter -VMName "{}" -SwitchName "{}"'.format(vm_name,
                                                                                switch_name)
    rs = run_ps(ps_script)

    return rs


def run_ps(ps: str) -> Response:
    """
    Run powershell script on target machine.

    Args:
        ps: Powershell script to run.
    Returns:
        Response object containing stderr, stdout and exit_status.
    """
    func_d = {'ssh': run_cmd_ssh,
              'winrm': run_cmd_winrm}
    proto = config['protocol']
    encoded_ps = b64encode(ps.encode('utf_16_le')).decode('ascii')
    rs = func_d[proto]('powershell -encodedcommand {}'.format(encoded_ps))

    return rs


def run_cmd_ssh(cmd: str) -> Response:
    """
    Run batch script using ssh client.

    Args:
        cmd: batch script to run.
    Returns:
        Response object containing stderr, stdout and exit_status.
    """
    ssh_client = SSHClient()
    ssh_client.load_system_host_keys()
    ssh_client.set_missing_host_key_policy(AutoAddPolicy())
    ssh_client.connect(username=config['user'],
                       password=config['pass'],
                       hostname=config['host'],
                       port=int(config['ssh_port']),
                       allow_agent=False,
                       look_for_keys=False)

    cmd_st = ssh_client.exec_command(cmd)

    rs = namedtuple('Response', ['std_out', 'std_err', 'status_code'])
    rs.std_out = cmd_st[1].read()
    rs.std_err = cmd_st[2].read()
    rs.status_code = cmd_st[1].channel.recv_exit_status()
    ssh_client.close()

    return rs


def run_cmd_winrm(cmd: str) -> Response:
    """
    Run batch script using winrm client.

    Args:
        cmd: batch script to run.
    Returns:
        Response object containing stderr, stdout and exit_status.
    """
    client = Protocol(endpoint='http://{}:5985/wsman'.format(config['host']),
                      transport='ntlm',
                      username=r'{}\{}'.format(config['domain'],
                                               config['user']),
                      password=config['pass'],
                      server_cert_validation='ignore')

    shell_id = client.open_shell()
    command_id = client.run_command(shell_id, cmd)
    rs = Response(client.get_command_output(shell_id, command_id))
    client.cleanup_command(shell_id, command_id)
    client.close_shell(shell_id)

    return rs



#!/usr/bin/env python3
# coding: utf-8
from hypy.cli_main import cli
from hypy.cli_snap import create, delete, restore, snap, snaps
from hypy.cli_switch import switch, switches

cli.add_command(switch)
cli.add_command(switches)

cli.add_command(snap)
cli.add_command(snaps)
cli.add_command(create)
cli.add_command(restore)
cli.add_command(delete)

cli()



from time import sleep

import click
from hypy.modules import cache, config, hvclient, printer


@click.group('cli')
@click.option('--user', '-u', help='Username in hyper-v server')
@click.option('passw', '--pass', '-p', help='Password in hyper-v server')
@click.option('--domain', '-d', help='Domain name')
@click.option('--host', '-m', help='Hyper-V server hostname/ip address')
@click.option('--proto', '-t', help='Protocol to be used',
              type=click.Choice(['ssh', 'winrm']))
def cli(user, passw, domain, host, proto):
    """
    Multiplataform Hyper-V Manager using Python and FreeRDP
    """
    config.load(user, passw, domain, host, proto)
    hvclient.config = config.configuration
    cache.current_host = config.configuration['host']
    cache.sync_interval = config.configuration['sync_interval']


@cli.command("status", help='Show virtual machine current status')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.pass_context
def show_status(ctx, by_name, ident):
    name = cache.get_name(by_name, ident)
    rs = hvclient.get_vm(name)
    vms = hvclient.parse_result(rs)
    cache.update_cache(vms)
    cache_vms = cache.list_vms()
    printer.print_list_vms(cache_vms, name)
    rs_snaps = hvclient.list_vm_snaps(name)
    snaps = hvclient.parse_result(rs_snaps)
    printer.print_vm_snaps(snaps, name, vms['ParentSnapshotName'])


@cli.command("list", help='List virtual machines and its indexes')
@click.option('--sync', '-s', is_flag=True, default=False,
              help='Syncronize with server updating local cache')
@click.option('--name', '-n', help='Filter virtual machines by name')
@click.option('--rem', '-r', is_flag=True, default=False, help='Remove old cache before sync')
def list_vms(sync, name, rem):
    remove_old_cache = rem or cache.need_update()
    if sync or remove_old_cache:
        rs = hvclient.get_vm(name)
        vms = hvclient.parse_result(rs)
        if remove_old_cache:
            cache.remove_cache()
        cache.update_cache(vms)
    cache_vms = cache.list_vms()
    printer.print_list_vms(cache_vms, name)


@cli.command("ls", help='List updated virtual machines and its indexes')
@click.option('--name', '-n', help='Filter virtual machines by name')
@click.option('--rem', '-r', is_flag=True, default=False, help='Remove old cache before sync')
@click.pass_context
def ls(ctx, name, rem):
    ctx.invoke(list_vms, sync=True, name=name, rem=rem)


@cli.command(help="Connect to virtual machine identified by index")
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
def connect(by_name, ident):
    name = cache.get_name(by_name, ident)
    vm_cache = cache.get_vm_by_name(name)

    vm_name = vm_cache['Name']
    vm_index = vm_cache['index']
    vm_id = vm_cache['Id']

    rs = hvclient.get_vm(vm_name)
    vm = hvclient.parse_result(rs)
    cache.update_cache(vm)

    if vm['State'] not in [2, 9]:
        rs = hvclient.start_vm(vm_name)
        hvclient.parse_result(rs)
        sleep(5)

    hvclient.connect(vm_id, vm_name, vm_index)


@cli.command(help='Start virtual machine identified by index')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.pass_context
def start(ctx, by_name, ident):
    name = cache.get_name(by_name, ident)
    ctx.invoke(list_vms, sync=False, name=name)
    rs = hvclient.start_vm(name)
    hvclient.parse_result(rs)
    ctx.invoke(list_vms, sync=True, name=name)


@cli.command(help='Pause virtual machine identified by index')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.pass_context
def pause(ctx, by_name, ident):
    name = cache.get_name(by_name, ident)
    ctx.invoke(list_vms, sync=False, name=name)
    rs = hvclient.pause_vm(name)
    hvclient.parse_result(rs)
    ctx.invoke(list_vms, sync=True, name=name)


@cli.command(help='Resume (paused) virtual machine identified by index')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.pass_context
def resume(ctx, by_name, ident):
    name = cache.get_name(by_name, ident)
    ctx.invoke(list_vms, sync=False, name=name)
    rs = hvclient.resume_vm(name)
    hvclient.parse_result(rs)
    ctx.invoke(list_vms, sync=True, name=name)


@cli.command(help='Stop virtual machine identified by index')
@click.option('--force', '-f', is_flag=True, help='Hyper-V gives the guest\
 five minutes to save data, then forces a shutdown')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.pass_context
def stop(ctx, by_name, ident, force):
    name = cache.get_name(by_name, ident)
    ctx.invoke(list_vms, sync=False, name=name)
    rs = hvclient.stop_vm(name, force)
    hvclient.parse_result(rs)
    ctx.invoke(list_vms, sync=True, name=name)


@cli.command(help='Save virtual machine identified by index')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.pass_context
def save(ctx, by_name, ident):
    name = cache.get_name(by_name, ident)
    ctx.invoke(list_vms, sync=False, name=name)
    rs = hvclient.save_vm(name)
    hvclient.parse_result(rs)
    ctx.invoke(list_vms, sync=True, name=name)



import click
from hypy.modules import cache, hvclient, printer


@click.group('snap')
def snap():
    """Manage virtual machine snapshots"""
    pass


@snap.command('ls', help='List virtual machine snapshots')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
def snap_ls(by_name, ident):
    name = cache.get_name(by_name, ident)
    rs = hvclient.get_vm(name)
    vms = hvclient.parse_result(rs)
    cache.update_cache(vms)
    cache_vms = cache.list_vms()
    printer.print_list_vms(cache_vms, name)
    rs_snaps = hvclient.list_vm_snaps(name)
    snaps = hvclient.parse_result(rs_snaps)
    printer.print_vm_snaps(snaps, name, vms['ParentSnapshotName'])


@snap.command('create', help="Create a new snapshot with vm's current state")
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.option('--snap_type',
              '-s',
              help='Snapshot type to be created',
              type=click.Choice(['standard', 'production']),
              default='standard')
@click.argument('snap_name')
def snap_create(by_name, ident, snap_name, snap_type):
    name = cache.get_name(by_name, ident)

    rs = hvclient.get_snapsshot_type(name)
    current_snap_type = hvclient.parse_result(rs)["CheckpointType"]

    if current_snap_type != hvclient.SNAP_TYPES[snap_type]:
        rs = hvclient.set_snapshot_type(name, snap_type)
        hvclient.parse_result(rs)

    rs = hvclient.create_vm_snapshot(name, snap_name)
    hvclient.parse_result(rs)


@snap.command('restore', help='Restore virtual machine snapshot')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.argument('snap_name')
def snap_restore(by_name, ident, snap_name):
    name = cache.get_name(by_name, ident)
    rs = hvclient.restore_vm_snap(name, snap_name)
    hvclient.parse_result(rs)


@snap.command('rm', help="Delete a machine's snapshot by name")
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.option('-r', is_flag=True, help="Remove snapshot's children as well")
@click.argument('snap_name')
def snap_delete(by_name, r, ident, snap_name):
    name = cache.get_name(by_name, ident)
    rs = hvclient.remove_vm_snapshot(name, snap_name, r)
    hvclient.parse_result(rs)


#
# Aliases
#
@click.command('snaps', help='List virtual machine snapshots')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.pass_context
def snaps(ctx, by_name, ident):
    ctx.forward(snap_ls)


@click.command('create', help="Create a new snapshot with vm's current state")
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.option('--snap_type',
              '-s',
              help='Snapshot type to be created',
              type=click.Choice(['standard', 'production']),
              default='standard')
@click.argument('snap_name')
@click.pass_context
def create(ctx, by_name, ident, snap_name, snap_type):
    ctx.forward(snap_create)


@click.command('restore', help='Restore virtual machine snapshot')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.argument('snap_name')
@click.pass_context
def restore(ctx, by_name, ident, snap_name):
    ctx.forward(snap_restore)


@click.command('delete', help="Delete a machine's snapshot by name")
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.option('-r', is_flag=True, help="Remove snapshot's children as well")
@click.argument('snap_name')
@click.pass_context
def delete(ctx, by_name, r, ident, snap_name):
    ctx.forward(snap_delete)



"""
Printer module. Formats messages to be printed in command line.
"""
from datetime import timedelta
from fnmatch import fnmatch

from hypy.modules.snaptree import create_tree

STATES = {3: 'off',
          2: 'running',
          9: 'paused',
          6: 'saved'}
ADJ = {'index': 3,
       'state': 7,
       'name': 30}


def print_vm_switch(switch_json: dict):
    """
    Print virtual machine's current virtual network switch.

    Args:
        switch_json: Dict containing current switch information.
    """
    if isinstance(switch_json, dict):
        switch_json = [switch_json]

    print("{} {}".format("VMName".ljust(ADJ['name']), "SwitchName"))
    for switch in switch_json:
        print("{} {}".format(str(switch['VMName']).ljust(ADJ['name']), switch['SwitchName']))


def print_switches(switches_json: dict):
    """
    Print a list of virtual network switches.

    Args:
        switches_json: Dict containing the table of switches.
    """
    print("-- Virtual network switches --")

    # Listing
    for switch in switches_json:
        print(switch['Name'])


def print_vm_snaps(snaps_json: dict, vm_name: str, current_snap: str):
    """
    Print ascii tree of checkpoints.

    Args:
        snaps_json: Dict containing the table of checkpoints.
        vm_name: Vm name to be shown as root of the tree.
    """
    if snaps_json:
        # If there is only one element, make it a list
        if isinstance(snaps_json, dict):
            snaps_json = [snaps_json]

        t_snaps = create_tree(snaps_json,
                              vm_name,
                              mark=current_snap,
                              f_pid="ParentSnapshotId",
                              f_id="Id",
                              f_label="Name",
                              f_ctime="CreationTime",
                              v_none=None)
        print("-- Virtual Machine Snapshots --")
        print(t_snaps)
    else:
        print("{} has no snapshots".format(vm_name))


def print_list_vms(vms_json: dict, filter_vms: str):
    """
    Print list of virtual machines.

    Args:
        vms_json: Dict containing the table of vms.
        filter_vms: Filter to be applied at the output. Only the vms whose name
            matches the filter will be shown.
    """
    # Listing
    # print("-- Hyper-V Virtual Machine Listing --")

    # Header
    print("{} {} {} {}".format("Index".rjust(ADJ['index']),
                               "State".ljust(ADJ['state']),
                               "Name".ljust(ADJ['name']),
                               "Uptime"))

    if filter_vms:
        vms_show = [vm for vm in vms_json if fnmatch(vm['Name'], filter_vms)]
    else:
        vms_show = vms_json

    # Listing
    for vm in vms_show:
        index = str(vms_json.index(vm)).rjust(ADJ['index'])
        state = STATES.get(vm['State'], "unknown").ljust(ADJ['state'])
        name = str(vm['Name']).ljust(ADJ['name'])
        uptime = str(timedelta(hours=vm['Uptime']['TotalHours']))
        print("[{}] {} {} {}".format(index, state, name, uptime))



"""
Hypy configuration manager module
"""
import configparser
from os.path import expanduser

CONFIG_FILE_LOCATION = expanduser('~/.hypy.conf')
configuration = None


def load(user: str, passw: str, domain: str, host: str, proto: str):
    """
    Read config file and command line to create the configuration.

    Args:
        user: User passed as command line option.
        passw: Password passed as command line option.
        domain: Domain passed as command line option.
        host: Host passed as command line option.
        proto: Protocol passed as command line option.
    """
    global configuration

    try:
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE_LOCATION)

        credentials = config['credentials']

        configuration = {'user': credentials['user'],
                         'pass': credentials['pass'],
                         'domain': credentials['domain'],
                         'host': credentials['host']}

        if user is not None:
            configuration['user'] = user
        if passw is not None:
            configuration['pass'] = passw
        if domain is not None:
            configuration['domain'] = domain
        if host is not None:
            configuration['host'] = host

        options = config['options']

        configuration['sync_interval'] = options['sync_interval']
        configuration['protocol'] = options['protocol']
        configuration['ssh_port'] = options['ssh_port']

        if proto is not None:
            configuration['protocol'] = proto

    except KeyError:
        print("Please, configure your credentials file - hypy.conf")
        print(CONFIG_FILE_LOCATION)
        exit(1)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hypy', 'hypy.modules']

package_data = \
{'': ['*']}

install_requires = \
['asciitree>=0.3.3,<0.4.0',
 'click>=7.0,<8.0',
 'paramiko>=2.6,<3.0',
 'pywinrm>=0.3.0,<0.4.0']

entry_points = \
{'console_scripts': ['hypy = hypy.__main__:cli']}

setup_kwargs = {
    'name': 'hypy3',
    'version': '0.3.6',
    'description': 'Multiplataform Hyper-V Manager using Python and FreeRDP',
    'long_description': None,
    'author': 'Gabriel Avanzi',
    'author_email': 'gabriel.avanzi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



import click
from hypy.modules import cache, hvclient, printer


@click.group('switch')
def switch():
    """Manage virtual network switches in the Hyper-V server."""
    pass


@switch.command('ls', help='List avaiable virtual network switches in the Hyper-V server')
def list_switches():
    rs = hvclient.list_switches()
    switches = hvclient.parse_result(rs)
    printer.print_switches(switches)


@switch.command('set', help='Change current vm network switch')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
@click.argument('switch_name')
def set_switch(by_name, ident, switch_name):
    name = cache.get_name(by_name, ident)
    rs = hvclient.set_switch(name, switch_name)
    hvclient.parse_result(rs)


@switch.command('get', help='Get current vm network switch')
@click.option('--name', '-n', 'by_name', is_flag=True, default=False,
              help='Use vm name instead of index')
@click.argument('ident')
def get_switch(by_name, ident):
    name = cache.get_name(by_name, ident)
    rs = hvclient.get_switch(name)
    switch = hvclient.parse_result(rs)
    printer.print_vm_switch(switch)


#
# Aliases
#
@click.command('switches', help='List avaiable virtual network switches in the Hyper-V server')
@click.pass_context
def switches(ctx):
    ctx.forward(list_switches)
